//#define BOOST_REGEX_NO_LIB
#define BOOST_REGEX_STATIC_LINK

#include <windows.h>
#include "../ls-b247/lsapi/lsapi.h"
#include "exports.h"

#pragma warning(push)
#pragma warning(disable: 4097 4127 4702) // for warning level 4
#include <string>
#include <utility>
#include <vector>
#include <boost/regex.hpp>
#include <boost/shared_ptr.hpp>
#pragma warning(pop)

using namespace boost;
using std::string;


#define CRIPPLED

#if !defined(CRIPPLED) && !defined(LOG_ERROR)
# define LOG_ERROR 1
# define LOG_WARNING 2
#endif

#define DUMB_NUMBER	    255
#define MB_NASTY        MB_OK | MB_ICONEXCLAMATION | MB_SETFOREGROUND

// not using map or set to avoid sorting
typedef std::pair<string, string> RegExp;
typedef std::vector<RegExp> RegExpressions;

RegExpressions g_RegExps;
const char g_szAppName[] = "fnxl.dll";
HWND g_hWnd = NULL;

UINT g_uMsgs[] =
{
    LM_WINDOWCREATED, LM_WINDOWACTIVATED, LM_GETMINRECT, LM_REDRAW,
    LM_TASKMAN, LM_REFRESH, 0
};

void LoadSettings();
void ChangeTitle(HWND hwnd);
void BangReloadConfig(HWND hCaller, const char* pszArgs);
BOOL CALLBACK EnumWindowsProc(HWND hwnd, LPARAM lParam);
LRESULT CALLBACK TheWndProc(HWND hWnd, UINT iMsg, WPARAM wParam, LPARAM lParam);

typedef shared_ptr<FILE> AutoFile;


// ---[ LoadSettings ]----------------------------------------------------------
//
void LoadSettings()
{
	g_RegExps.clear();

    char szConfigFile[MAX_PATH] = { 0 };
	
	// only open if the user has given us a file.
	if (GetRCString("FNXLConfigFile", szConfigFile, NULL, MAX_PATH))
    {
		FILE* f = fopen(szConfigFile, "rt");
		AutoFile autofile(f, fclose);

		if (f)
        {
			// loop through file reading two lines at a time
			while ( !feof(f) )
            {
				char szRegExp[DUMB_NUMBER];
				char szPattern[DUMB_NUMBER];

                // read the first line (regexp)
				if ( fgets(szRegExp, DUMB_NUMBER, f))
                {
                    // read the second line (pattern)
					if ( fgets(szPattern, DUMB_NUMBER, f) )
                    {
						int iLen = strlen(szRegExp);

						if (iLen && szRegExp[iLen-1] == '\n')
                        {
                            szRegExp[iLen-1] = '\0';
                        }

						iLen = strlen(szPattern);

						if (iLen && szPattern[iLen-1] == '\n')
                        {
                            szPattern[iLen-1] = '\0';
                        }

						if ( szRegExp[0] != '\0' && szPattern != '\0' )
                        {
                            g_RegExps.push_back(RegExp(szRegExp, szPattern));
						}
					}

				}

			}
		}
		else
		{
			#ifndef CRIPPLED
			LSLogPrintf(LOG_ERROR, g_szAppName, "Unable to open FNXLConfigFile: %s", szConfigFile);
			#endif
		}

	}
}

// ---[ ChangeTitle ]-----------------------------------------------------------
//
void ChangeTitle(HWND hwnd)
{
	if ( !IsWindow(hwnd) )
		return;
	
    char buffer[DUMB_NUMBER];
	GetWindowText(hwnd, buffer, DUMB_NUMBER);

    string title = buffer;

    for (RegExpressions::iterator itExp = g_RegExps.begin();
         itExp != g_RegExps.end(); ++itExp)
	{
		regex rePattern(itExp->first);
	
        if (regex_match(title, rePattern))
        {
            string result = regex_merge(title, rePattern, itExp->second,
                match_default | format_no_copy);
            
            if (!result.empty())
            {
                SetWindowText(hwnd, result.c_str());
            }
        }
        
        #ifndef CRIPPLED
        LSLogPrintf(LOG_ERROR, g_szAppName, "An error occured in one of the following regexps: %s  %s", itExp->first, itExp->second);
        #endif
	}
}

// ---[ BangReloadConfig ]------------------------------------------------------
//
void BangReloadConfig(HWND /* hCaller */, const char* /* pszArgs */)
{
	LoadSettings();
	
    if (!g_RegExps.empty())
    {
        EnumWindows(EnumWindowsProc, 0);
    }
}


// ---[ EnumWindowsProc ]-------------------------------------------------------
//
BOOL CALLBACK EnumWindowsProc(HWND hwnd, LPARAM /* lParam */)
{
	ChangeTitle(hwnd);
	return TRUE;
}


// ---[ TheWndProc ]------------------------------------------------------------
//
LRESULT CALLBACK TheWndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch ( uMsg )
	{
    case LM_REFRESH:
        BangReloadConfig(NULL, NULL);
        break;
    case LM_WINDOWCREATED:
	case LM_WINDOWACTIVATED:
	case LM_GETMINRECT:
	case LM_REDRAW:
	case LM_TASKMAN:
		ChangeTitle((HWND)wParam);
		return 0;
		break;
	case WM_CREATE:
		{
			EnumWindows(EnumWindowsProc, 0);
			SendMessage(GetLitestepWnd(), LM_REGISTERMESSAGE, (WPARAM)hWnd, (LPARAM)g_uMsgs);
		}
		return 0;
		break;
	case WM_DESTROY:
			SendMessage(GetLitestepWnd(), LM_UNREGISTERMESSAGE, (WPARAM)hWnd, (LPARAM)g_uMsgs);
		break;
	}
	return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

// ---[ initModuleEx ]----------------------------------------------------------
//
__declspec( dllexport ) int initModuleEx(HWND /* hLitestep */, HINSTANCE hInstance, LPCSTR /* pszPath */)
{
	LoadSettings();

	if (!g_RegExps.empty())
	{
		WNDCLASS wc;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = 0;
		wc.hbrBackground = 0;
		wc.hCursor = 0;
		wc.hIcon = 0;
		wc.hInstance = hInstance;
		wc.lpfnWndProc = TheWndProc;
		wc.lpszClassName = g_szAppName;
		wc.lpszMenuName = 0;
		wc.style = 0;

		if ( !RegisterClass(&wc) )
        {
			// try again
			UnregisterClass(g_szAppName, hInstance);
			if ( !RegisterClass(&wc) )
            {
				#ifndef CRIPPLED
				LSLogPrintf(LOG_ERROR, g_szAppName, "Unable to register window class");
				#else
				MessageBox(0, "Unable to register window class.", g_szAppName, MB_NASTY);
				#endif
				return -1;
			}
		}

		g_hWnd = CreateWindow(g_szAppName,"dead window",WS_POPUP,0,0,0,0,NULL,NULL, hInstance,0);
		if (!g_hWnd)
        {
			#ifndef CRIPPLED
			LSLogPrintf(LOG_ERROR, g_szAppName, "Unable to create window");
			#else
			MessageBox(0, "Unable to create window.", g_szAppName, MB_NASTY);
			#endif
			return -1;
		}
		AddBangCommand("!FnXLReloadConfig", BangReloadConfig);
	}
	else
	{
		#ifndef CRIPPLED
		LSLogPrintf(LOG_WARNING, g_szAppName, "No config lines were found");
        #endif
	}

	return 0;
}

// ---[ quitModule ]------------------------------------------------------------
//
__declspec( dllexport ) int quitModule(HINSTANCE hInstance)
{
	if (g_hWnd)
    {
		DestroyWindow(g_hWnd);
		RemoveBangCommand("!FnXLReloadConfig");
	}
	
	UnregisterClass(g_szAppName, hInstance);

	g_RegExps.clear();

	return 0;
}
